/**
 * @licstart The following is the entire license notice for the
 * Javascript code in this page
 *
 * Copyright 2018 Mozilla Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * @licend The above is the entire license notice for the
 * Javascript code in this page
 */
"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ScaleJS = void 0;

var _util = require("../shared/util");

var ScaleJS = function ScaleJSClosure() {
  function ScaleJS() {}

  ScaleJS.prototype = {
    determineScale: function determineScale(width, height) {
      var testColor = '#ffffff';
      var colorData = new Uint8ClampedArray(4);
      var scale = 0;
      var canvas = document.createElement('canvas');
      var context = canvas.getContext('2d');

      while (colorData[0] !== 255) {
        scale += 1;
        canvas.width = width / scale;
        canvas.height = height / scale;
        context.fillStyle = testColor;
        context.fillRect(0, 0, 1, 1);
        colorData = context.getImageData(0, 0, 1, 1).data;
      }

      return scale;
    },
    needsScale: function needsScale(width, height, colorSpace, bpc) {
      var scaleObj = {
        width: width,
        height: height,
        scale: 1,
        transformImage: false
      };
      var scale = this.determineScale(width, height);

      if (scale > 1) {
        scaleObj.transformImage = true;
        scaleObj.scale = scale;

        if (colorSpace === 'DeviceGray' && bpc === 1 && scaleObj.scale > 2) {
          scaleObj.scale = 4;
        }

        scaleObj.width = Math.ceil(width / scaleObj.scale);
        scaleObj.height = Math.ceil(height / scaleObj.scale);
      }

      return scaleObj;
    },
    scaleImg: function scaleImg(imgData) {
      if (!imgData.scaleProcessed) {
        var imgScaleObj;

        if (imgData.kind === _util.ImageKind.GRAYSCALE_1BPP) {
          imgScaleObj = this.needsScale(imgData.width, imgData.height, 'DeviceGray', 1);

          if (imgScaleObj.transformImage) {
            imgData.kind = _util.ImageKind.GRAYSCALE_8BPP;
            imgData.data = this.scaleBWImage(imgData.data, imgData.width, imgData.height, imgData.width + 7 >> 3, imgScaleObj.scale);
            imgData.width = imgScaleObj.width;
            imgData.height = imgScaleObj.height;
          }
        } else if (imgData.kind === _util.ImageKind.RGB_24BPP) {
          imgScaleObj = this.needsScale(imgData.width, imgData.height, 'RGB', 1);

          if (imgScaleObj.transformImage) {
            imgData.data = this.scaleRGBImage(imgData.data, imgData.width, imgData.height, imgScaleObj.scale);
            imgData.width = imgScaleObj.width;
            imgData.height = imgScaleObj.height;
          }
        }

        imgData.scaleProcessed = true;
      }

      return imgData;
    },
    scaleMask: function scaleMask(imgData) {
      if (!imgData.scaleProcessed) {
        imgData.isMaskScaled = false;
        var imgScaleObj = this.needsScale(imgData.width, imgData.height, 'DeviceGray', 1);

        if (imgScaleObj.transformImage) {
          imgData.data = this.scaleBWImage(imgData.data, imgData.width, imgData.height, imgData.width + 7 >> 3, imgScaleObj.scale);
          imgData.width = imgScaleObj.width;
          imgData.height = imgScaleObj.height;
          imgData.isMaskScaled = true;
        }

        imgData.scaleProcessed = true;
      }

      return imgData;
    },
    bitCount: function bitCount(n) {
      var count = 0;

      while (n) {
        count += n & 1;
        n >>= 1;
      }

      return count;
    },
    resizeOriginalImage: function resizeOriginalImage(originalImage, originalHeight, rowBytes) {
      var drawImage = new Uint8Array(originalHeight * rowBytes);
      drawImage.set(originalImage);
      drawImage.fill(255, originalImage.length, originalHeight * rowBytes);
      return drawImage;
    },
    scaleMainArea: function scaleMainArea(newImage, img, rowBytes, newRowBytes, stopRow, stopCol, scale) {
      var m1;
      var m2;
      var idx;
      var row, col;

      if (scale === 2) {
        m1 = [0, 64, 64, 128];
        m2 = [0, 64, 64, 127];

        for (row = 0; row < stopRow; row += scale) {
          for (col = 0; col < stopCol; col++) {
            idx = row * newRowBytes / 2 + col * 4;
            newImage[idx++] = m1[img[row * rowBytes + col] >> 6 & 0x03] + m2[img[(row + 1) * rowBytes + col] >> 6 & 0x03];
            newImage[idx++] = m1[img[row * rowBytes + col] >> 4 & 0x03] + m2[img[(row + 1) * rowBytes + col] >> 4 & 0x03];
            newImage[idx++] = m1[img[row * rowBytes + col] >> 2 & 0x03] + m2[img[(row + 1) * rowBytes + col] >> 2 & 0x03];
            newImage[idx++] = m1[img[row * rowBytes + col] >> 0 & 0x03] + m2[img[(row + 1) * rowBytes + col] >> 0 & 0x03];
          }
        }
      } else if (scale === 4) {
        m1 = [0, 16, 16, 32, 16, 32, 32, 48, 16, 32, 32, 48, 32, 48, 48, 64];
        m2 = [0, 16, 16, 32, 16, 32, 32, 48, 16, 32, 32, 48, 32, 48, 48, 63];

        for (row = 0; row < stopRow; row += scale) {
          for (col = 0; col < stopCol; col++) {
            idx = row * newRowBytes / 4 + col * 2;
            newImage[idx++] = m1[img[row * rowBytes + col] >> 4 & 0x0F] + m1[img[(row + 1) * rowBytes + col] >> 4 & 0x0F] + m1[img[(row + 2) * rowBytes + col] >> 4 & 0x0F] + m2[img[(row + 3) * rowBytes + col] >> 4 & 0x0F];
            newImage[idx++] = m1[img[row * rowBytes + col] >> 0 & 0x0F] + m1[img[(row + 1) * rowBytes + col] >> 0 & 0x0F] + m1[img[(row + 2) * rowBytes + col] >> 0 & 0x0F] + m2[img[(row + 3) * rowBytes + col] >> 0 & 0x0F];
          }
        }
      }
    },
    scaleOtherAreas: function scaleOtherAreas(newImage, img, rowBytes, newRowBytes, startRow, stopRow, startCol, stopCol, offsetX, offsetY, scale) {
      var mask = 0x03;

      if (scale === 4) {
        mask = 0x0F;
      }

      var multFactor1 = 256 / (scale * offsetY);
      var multFactor2 = 255 / (scale * offsetY);
      var lastColPixelGroups = Math.ceil(offsetX / scale);
      var offsetXMod = offsetX % scale;
      var offsetXDiv = Math.floor(offsetX / scale);
      var shiftValue, pixValue;

      for (var row = startRow; row < stopRow; row += scale) {
        for (var col = startCol; col < stopCol; col++) {
          var idx = row * newRowBytes / scale + col * (8 / scale);

          for (var colGroup = 0; colGroup < lastColPixelGroups; colGroup++) {
            shiftValue = 8 - scale - scale * colGroup;

            if (offsetXMod !== 0 && offsetXDiv === colGroup) {
              if (scale === 2) {
                mask = 0x02;
              } else {
                switch (offsetXMod) {
                  case 3:
                    mask = 0x0E;
                    break;

                  case 2:
                    mask = 0x0C;
                    break;

                  case 1:
                    mask = 0x08;
                    break;
                }
              }

              multFactor1 = 256 / (offsetXMod * offsetY);
              multFactor2 = 255 / (offsetXMod * offsetY);
            }

            pixValue = 0;

            for (var j = 0; j < offsetY; j++) {
              if (j < offsetY - 1) {
                pixValue += this.bitCount(img[(row + j) * rowBytes + col] >> shiftValue & mask) * multFactor1;
              } else {
                pixValue += this.bitCount(img[(row + j) * rowBytes + col] >> shiftValue & mask) * multFactor2;
              }
            }

            newImage[idx++] = pixValue;
          }
        }
      }
    },
    scaleBWImage: function scaleBWImage(img, originalWidth, originalHeight, rowBytes, scale) {
      var image = img;

      if (img.length < originalHeight * rowBytes) {
        image = this.resizeOriginalImage(img, originalHeight, rowBytes);
      }

      if (scale !== 2 && scale !== 4) {
        scale = scale > 2 ? 4 : 2;
      }

      var newRowBytes = Math.ceil(originalWidth / scale);
      var newHeight = Math.ceil(originalHeight / scale);
      var newImage = new Uint8Array(newHeight * newRowBytes);
      var offsetX = originalWidth % 8;
      var offsetY = originalHeight % scale;
      var stopRow = originalHeight - offsetY;
      var stopCol = rowBytes - (offsetX === 0 ? 0 : 1);
      this.scaleMainArea(newImage, image, rowBytes, newRowBytes, stopRow, stopCol, scale);

      if (offsetX !== 0) {
        this.scaleOtherAreas(newImage, image, rowBytes, newRowBytes, 0, stopRow, rowBytes - 1, rowBytes, offsetX, scale, scale);
      }

      if (offsetY !== 0) {
        this.scaleOtherAreas(newImage, image, rowBytes, newRowBytes, originalHeight - offsetY, originalHeight, 0, stopCol, 8, offsetY, scale);
      }

      if (offsetX !== 0 && offsetY !== 0) {
        this.scaleOtherAreas(newImage, image, rowBytes, newRowBytes, originalHeight - offsetY, originalHeight, rowBytes - 1, rowBytes, offsetX, offsetY, scale);
      }

      return newImage;
    },
    filterArea: function filterArea(origImg, filteredImg, scale, rowBytes, newRowBytes, numComps, filterData) {
      var sumR, sumG, sumB;
      var boxRow, boxCol;
      var destIdx = 0,
          srcCol,
          srcRowIdx,
          srcPixelIdx;
      var scaleByScale = filterData.rowWindow * filterData.colWindow;

      for (var row = filterData.startRow; row < filterData.endRow; row++) {
        var srcRow = row * scale;

        for (var col = filterData.startCol; col < filterData.endCol; col += numComps) {
          destIdx = row * newRowBytes + col;
          srcCol = col * scale;
          sumR = 0;
          sumG = 0;
          sumB = 0;

          for (boxRow = 0; boxRow < filterData.rowWindow; boxRow++) {
            srcRowIdx = srcCol + rowBytes * (srcRow + boxRow);

            for (boxCol = 0; boxCol < filterData.colWindow; boxCol++) {
              srcPixelIdx = srcRowIdx + numComps * boxCol;
              sumR += origImg[srcPixelIdx];
              sumG += origImg[srcPixelIdx + 1];
              sumB += origImg[srcPixelIdx + 2];
            }
          }

          filteredImg[destIdx++] = sumR / scaleByScale;
          filteredImg[destIdx++] = sumG / scaleByScale;
          filteredImg[destIdx++] = sumB / scaleByScale;
        }
      }
    },
    scaleRGBImage: function scaleRGBImage(img, originalWidth, originalHeight, scale) {
      var drawWidth = Math.ceil(originalWidth / scale);
      var drawHeight = Math.ceil(originalHeight / scale);
      var numComps = 3;
      var rowBytes = originalWidth * numComps;
      var newRowBytes = drawWidth * numComps;
      var newImage = new Uint8Array(drawHeight * newRowBytes);
      var overflowRowCnt = originalHeight % scale;
      var overflowDestRow = overflowRowCnt > 0 ? drawHeight - 1 : originalHeight;
      var overflowColCnt = originalWidth % scale;
      var overflowDestCol = overflowColCnt > 0 ? newRowBytes - numComps : newRowBytes;
      this.filterArea(img, newImage, scale, rowBytes, newRowBytes, numComps, {
        startRow: 0,
        endRow: overflowDestRow,
        startCol: 0,
        endCol: overflowDestCol,
        rowWindow: scale,
        colWindow: scale
      });

      if (overflowRowCnt !== 0) {
        this.filterArea(img, newImage, scale, rowBytes, newRowBytes, numComps, {
          startRow: overflowDestRow,
          endRow: drawHeight,
          startCol: 0,
          endCol: overflowDestCol,
          rowWindow: overflowRowCnt,
          colWindow: scale
        });
      }

      if (overflowColCnt !== 0) {
        this.filterArea(img, newImage, scale, rowBytes, newRowBytes, numComps, {
          startRow: 0,
          endRow: overflowDestRow,
          startCol: overflowDestCol,
          endCol: newRowBytes,
          rowWindow: scale,
          colWindow: overflowColCnt
        });
      }

      if (overflowRowCnt !== 0 && overflowColCnt !== 0) {
        this.filterArea(img, newImage, scale, rowBytes, newRowBytes, numComps, {
          startRow: overflowDestRow,
          endRow: drawHeight,
          startCol: overflowDestCol,
          endCol: newRowBytes,
          rowWindow: overflowRowCnt,
          colWindow: overflowColCnt
        });
      }

      return newImage;
    }
  };
  return ScaleJS;
}();

exports.ScaleJS = ScaleJS;